import re
from dataclasses import dataclass, field
from textwrap import wrap
from typing import List, Literal, Optional, Tuple

import botocore.model
from markdownify import markdownify


@dataclass
class FormatterArgs:
    line_range: Optional[Tuple[int, int]] = None
    length_range: Optional[Tuple[int, int]] = None
    black: bool = True
    style: Literal["sphinx", "epytext"] = "sphinx"
    force_wrap: bool = False
    make_summary_multi_line: bool = True
    pre_summary_newline: bool = True
    post_summary_newline: bool = True
    post_description_blank: bool = False
    non_strict: bool = False
    rest_section_adorns: str = r"""[!\"#$%&'()*+,-./\\:;<=>?@[]^_`{|}~]{4,}"""
    tab_width: int = 4
    wrap_summaries: int = 79
    wrap_descriptions: int = 79
    non_cap: List[str] = field(default_factory=list)


class DocumentationFormatter:
    MARKDOWN_LINK_RE = re.compile(
        r"(?:\[(?P<text>.*?)\])\((?P<link>.*?)\)", re.MULTILINE | re.DOTALL
    )
    PY_OBJECT_RE = re.compile(r"py:(.*?):``(.*?)``", re.MULTILINE | re.DOTALL)

    def __init__(self, max_length: int = 79):
        #: Wrap lines at this length.
        self.max_length = max_length

    def _clean_uls(self, documentation: str) -> str:
        """
        Look through ``documentation`` for unordered lists and clean them up.

        This means wrapping them properly at 79 characters, and adding a blank
        line before and after.

        Args:
            documentation: the partially processed reStructuredText
                documentation

        Returns:
            The documentation with unordered lists cleaned up.

        """
        lines = []
        source_lines = documentation.split("\n")
        for i, line in enumerate(source_lines):
            if line.startswith("*"):
                previous_line = source_lines[i - 1]
                if previous_line.strip() != "" and not previous_line.startswith("*"):
                    lines.append("")
                if len(line) > self.max_length:
                    wrapped = wrap(line, self.max_length)
                    lines.append(wrapped[0])
                    lines.extend([f"  {line}" for line in wrapped[1:]])
                else:
                    lines.append(line)
            elif len(line) > self.max_length:
                lines.extend(wrap(line, self.max_length))
            else:
                lines.append(line)
        return "\n".join(lines)

    def _clean_links(self, documentation: str) -> str:
        """
        Transform our Markdown links to reStructuredText links.

        Args:
            documentation: the partially processed reStructuredText
                documentation

        Returns:
            The documentation with links cleaned up.

        """
        for match in self.MARKDOWN_LINK_RE.finditer(documentation):
            text = match.group("text")
            link = match.group("link")
            link = link.replace(" ", "")
            documentation = documentation.replace(match.group(0), f"`{text} <{link}>`_")
        return documentation

    def _undo_double_backticks(self, documentation: str) -> str:
        """
        If we have custom docstrings that are already in reStructuredText, then
        we can end up with double backticks in our documentation from when we
        convert single back ticks to double in the Markdown -> reStructuredText
        conversion.  We need to undo some of those, especially when we have
        ``:py:obj:`` style references.

        Args:
            documentation: input documentation

        Returns:
            Cleaned up documentation.

        """
        for match in self.PY_OBJECT_RE.finditer(documentation):
            py_obj = match.group(0)
            updated_py_obj = py_obj.replace("``", "`")
            documentation = documentation.replace(py_obj, updated_py_obj)
        return documentation

    def clean(self, documentation: str, max_lines: Optional[int] = None) -> str:
        """
        Take the input documentation in HTML format and clean it up for use in a
        docstring, as reStructuredText.

        Args:
            documentation: the HTML documentation to clean up

        Keyword Args:
            max_lines: the maximum number of lines to include in the output

        Returns:
            Properly formatted reStructuredText documentation.

        """
        documentation = markdownify(documentation)
        if max_lines is not None:
            documentation = "\n".join(documentation.split("\n")[:max_lines])
        if "\n" in documentation:
            documentation = "\n".join(
                [line.strip() for line in documentation.split("\n")]
            )
        documentation = documentation.replace("`", "``")
        documentation = self._clean_uls(documentation)
        documentation = self._clean_links(documentation)
        documentation = self._undo_double_backticks(documentation)
        # remove any double backticks
        documentation = re.sub(r"\\{1}", "", documentation)
        # Change en-dashes to hyphens
        documentation = documentation.replace("–", "-")  # noqa: RUF001
        # Change forward ticks to backticks
        return documentation.replace("‘", "`")  # noqa: RUF001

    def format_docstring(self, shape: botocore.model.Shape) -> str:
        """
        Format the documentation for a model.

        Args:
            shape: the botocore shape for the model

        Returns:
            The formatted documentation for the model as reStructuredText.

        """
        documentation = shape.documentation
        documentation = self.clean(documentation)
        return documentation  # noqa: RET504

    def format_attribute(self, docs: str) -> str:
        """
        Format the documentation for a single attribute of a model.

        Args:
            docs: the documentation for the attribute

        Returns:
            The formatted documentation for the attribute as reStructuredText.

        """
        documentation = self.clean(docs, max_lines=1)
        docs = '    """\n'
        docs += documentation
        docs += '\n    """'
        return docs

    def format_argument(self, arg: str, docs: Optional[str]) -> str:
        """
        Format the documentation for a single argument of a method.

        Args:
            arg: the name of the argument
            docs: the documentation for the argument

        Returns:
            The formatted documentation for the argument as reStructuredText.

        """
        if not docs:
            docs = f"the value to set for {arg}"
        documentation = self.clean(docs, max_lines=1)
        lines = wrap(f"{arg}: {documentation}", self.max_length - 4)
        lines[0] = f"            {lines[0]}"
        for i, line in enumerate(lines[1:]):
            lines[i + 1] = f"                {line}"
        lines[-1] += "\n"
        return "\n".join([line.rstrip() for line in lines])



import re
from typing import TYPE_CHECKING, Callable, List, Optional, cast

if TYPE_CHECKING:
    from botocraft.services import (
        Cluster,
        ContainerInstance,
        Service,
        Task,
        TaskDefinition,
        TaskManager,
    )


# ---------
# Functions
# ---------


def extract_task_family_and_revision(task_definition_arn: str) -> str:
    """
    Extract the task family and revision from a task definition ARN.

    Args:
        task_definition_arn: The ARN of the task definition.

    Returns:
        The task family and revision in the format ``<family>:<revision>``.

    """
    task_definition_arn_re = r"arn:aws:ecs:[^:]+:[^:]+:task-definition/(?P<family>[^:]+):(?P<revision>[0-9]+)"  # noqa: E501
    match = re.match(task_definition_arn_re, task_definition_arn)
    assert (
        match
    ), f"Could not extract task family and revision from {task_definition_arn}"
    return f'{match.group("family")}:{match.group("revision")}'


# ----------
# Decorators
# ----------


# Service


def ecs_services_only(func: Callable[..., List[str]]) -> Callable[..., List["Service"]]:
    """
    Wraps :py:meth:`botocraft.services.ecs.ServiceManager.list` to return a list of
    :py:class:`botocraft.services.ecs.Service` objects instead of only a list of
    ARNs.
    """

    def wrapper(self, *args, **kwargs) -> List["Service"]:
        arns = func(self, *args, **kwargs)
        services = []
        # We have to do this in batches of 10 because the get_many method,
        # which uses the boto3 ``describe_services`` method, only accepts 10 ARNs
        # at a time.
        for i in range(0, len(arns), 10):
            services.extend(
                self.get_many(
                    arns[i : i + 10], cluster=kwargs["cluster"], include=["TAGS"]
                )
            )
        return services

    return wrapper


# Cluster


def ecs_clusters_only(func: Callable[..., List[str]]) -> Callable[..., List["Cluster"]]:
    """
    Wraps :py:meth:`botocraft.services.ecs.ClusterManager.list` to return a list of
    :py:class:`botocraft.services.ecs.Cluster` objects instead of only a list of
    ARNs.
    """

    def wrapper(self, *args, **kwargs) -> List["Cluster"]:
        arns = func(self, *args, **kwargs)
        clusters = []
        # We have to do this in batches of 100 because the get_many method,
        # which uses the boto3 ``describe_clusters`` method, only accepts 100 ARNs
        # at a time.
        for i in range(0, len(arns), 100):
            clusters.extend(self.get_many(clusters=arns[i : i + 100]))
        return clusters

    return wrapper


# TaskDefinition


def ecs_task_definitions_only(
    func: Callable[..., List[str]],
) -> Callable[..., List["TaskDefinition"]]:
    """
    Decorator to convert a list of ECS task definition identifiers to a list of
    :py:class:`botocraft.services.ecs.TaskDefinition` objects.
    """

    def wrapper(self, *args, **kwargs) -> List["TaskDefinition"]:
        identifiers = func(self, *args, **kwargs)
        return [self.get(identifier, include=["TAGS"]) for identifier in identifiers]

    return wrapper


# ContainerInstance


def ecs_container_instances_only(
    func: Callable[..., List[str]],
) -> Callable[..., List["ContainerInstance"]]:
    """
    Decorator to convert a list of ECS container instance arns to a list of
    :py:class:`botocraft.services.ecs.ContainerInstance` objects.
    """

    def wrapper(self, *args, **kwargs) -> List["ContainerInstance"]:
        arns = func(self, *args, **kwargs)
        container_instances = []
        for i in range(0, len(arns), 100):
            container_instances.extend(
                self.get_many(
                    cluster=kwargs["cluster"], containerInstances=arns[i : i + 100]
                )
            )
        return container_instances

    return wrapper


def ecs_container_instances_tasks_only(
    func: Callable[..., List[str]],
) -> Callable[..., List["Task"]]:
    """
    Decorator to convert a list of ECS container instance arns to a list of
    :py:class:`botocraft.services.ecs.ContainerInstance` objects.
    """

    def wrapper(self, *args, **kwargs) -> List["Task"]:
        from botocraft.services.ecs import Task

        arns = func(self, *args, **kwargs)
        tasks = []
        for i in range(0, len(arns), 100):
            tasks.extend(cast("TaskManager", Task.objects).get_many(arns[i : i + 100]))
        return tasks

    return wrapper


# Task
def ecs_task_populate_taskDefinition(  # noqa: N802
    func: Callable[..., Optional["Task"]],
) -> Callable[..., Optional["Task"]]:
    """
    Wraps :py:meth:`botocraft.services.ecs.TaskManager.get` to populate the
    :py:attr:`botocraft.services.ecs.Task.taskDefinition` attribute.

    We set the ``taskDefinition`` attribute to the task family and revision in the
    format ``<family>:<revision>``.  ``taskDefinition`` is an extra field that we
    add to the :py:class:`botocraft.services.ecs.Task` object that is not in the
    original botocore shape, but is useful for our purposes.
    """

    def wrapper(self, *args, **kwargs) -> Optional["Task"]:
        task = func(self, *args, **kwargs)
        if task:
            task.taskDefinition = extract_task_family_and_revision(
                task.taskDefinitionArn
            )
        return task

    return wrapper


def ecs_task_populate_taskDefinitions(  # noqa: N802
    func: Callable[..., List["Task"]],
) -> Callable[..., List["Task"]]:
    """
    Wraps :py:meth:`botocraft.services.ecs.TaskManager.get_many` to
    populate the :py:attr:`botocraft.services.ecs.Task.taskDefinition` attribute
    on each task.

    We set the ``taskDefinition`` attribute to the task family and revision in the
    format ``<family>:<revision>``.  ``taskDefinition`` is an extra field that we
    add to the :py:class:`botocraft.services.ecs.Task` object that is not in the
    original botocore shape, but is useful for our purposes.
    """

    def wrapper(self, *args, **kwargs) -> List["Task"]:
        tasks = func(self, *args, **kwargs)
        for task in tasks:
            task.taskDefinition = extract_task_family_and_revision(
                task.taskDefinitionArn
            )
        return tasks

    return wrapper


def ecs_tasks_only(func: Callable[..., List[str]]) -> Callable[..., List["Task"]]:
    """
    Wrap :py:meth:`botocraft.services.ecs.TaskManager.list` to return a list of
    :py:class:`botocraft.services.ecs.Task` objects instead of only a list of
    ARNs.
    """

    def wrapper(self, *args, **kwargs) -> List["Task"]:
        arns = func(self, *args, **kwargs)
        tasks = []
        # We have to do this in batches of 100 because the get_many method,
        # which uses the boto3 ``describe_tasks`` method, only accepts 100 ARNs
        # at a time.
        for i in range(0, len(arns), 100):
            tasks.extend(
                self.get_many(cluster=kwargs["cluster"], tasks=arns[i : i + 100])
            )
        return tasks

    return wrapper


# Mixins


class ECSServiceModelMixin:
    """
    A mixin for :py:class:`botocraft.services.ecs.Service` that adds
    some additional methods that we can't auto generate.
    """

    @property
    def required_cpu(self) -> int:
        """
        The required CPU for the service in CPU shares.  One full CPU is
        equivalent to 1024 CPU shares.
        """
        cpu: int = 0
        td = self.task_definition  # type: ignore[attr-defined]
        if td.cpu:
            cpu = int(td.cpu)
        else:
            for container in td.containerDefinitions:
                if container.cpu:
                    cpu += container.cpu
        return cpu

    @property
    def required_memory(self) -> int:
        """
        Return the required memory for the service in MiB.
        """
        memory: int = 0
        td = self.task_definition  # type: ignore[attr-defined]
        if td.memory:
            memory = int(td.memory)
        else:
            for container in td.containerDefinitions:
                if container.memory:
                    memory += container.memory
        return int(memory)

    @property
    def container_instances(self) -> List["ContainerInstance"]:
        """
        Return the :py:class:`botocraft.services.ecs.ContainerInstance` objects which
        are running our tasks for the service.
        """
        return [task.container_instance for task in self.tasks]  # type: ignore[attr-defined]

    @property
    def is_stable(self) -> bool:
        """
        Return whether the service is stable or not.
        """
        # this is the same test that the `services_stable` waiter uses
        return len(self.deployments) == 1 and (self.runningCount == self.desiredCount)  # type: ignore[attr-defined]

    def wait_until_stable(self, max_attempts: int = 40, delay: int = 15) -> None:
        """
        Wait until the service is stable.

        Raises:
            botocore.exceptions.WaiterError: if the service is not stable after
                ``max_attempts``, or some other error occurred.

        Keyword Args:
            max_attempts: The maximum number of attempts to make before giving
                up.
            delay: The number of seconds to wait between attempts.

        """
        waiter_config = {}
        if max_attempts:
            waiter_config["maxAttempts"] = max_attempts
        if delay:
            waiter_config["delay"] = delay
        if waiter_config:
            waiter_config["operation"] = "DescribeServices"  # type: ignore[assignment]
        waiter = self.objects.get_waiter("services_stable", WaiterConfig=waiter_config)  # type: ignore[attr-defined]
        waiter.wait(cluster=self.clusterArn, services=[self.serviceName])  # type: ignore[attr-defined]

    def scale(
        self,
        desired_count: int,
        wait: bool = False,
    ) -> None:
        """
        Scale the service to the desired count.  If ``wait`` is True, this will
        wait for the service to reach the desired count using the ``services_stable``
        boto3 waiter.

        Args:
            desired_count: The number of tasks to run.

        Keyword Args:
            wait: If True, wait for the service to reach the desired count.

        """
        self.objects.partial_update(  # type: ignore[attr-defined]
            self.serviceName,  # type: ignore[attr-defined]
            cluster=self.clusterArn,  # type: ignore[attr-defined]
            desiredCount=desired_count,
        )
        waiter = self.objects.get_waiter("services_stable")  # type: ignore[attr-defined]
        if wait:
            waiter.wait(
                cluster=self.clusterArn,  # type: ignore[attr-defined]
                services=[self.serviceName],  # type: ignore[attr-defined]
            )


class ECSContainerInstanceModelMixin:
    @property
    def free_cpu(self) -> int:
        """
        Return the free CPU shares on the container instance.  One full CPU is
        equivalent to 1024 CPU shares.
        """
        value: int = 0
        for resource in self.remainingResources:  # type: ignore[attr-defined]
            if resource.name == "CPU":
                value = int(resource.integerValue)
        return value

    @property
    def free_ram(self) -> int:
        """
        Return the free RAM in MiB on the container instance.
        """
        value: int = 0
        for resource in self.remainingResources:  # type: ignore[attr-defined]
            if resource.name == "MEMORY":
                value = int(resource.integerValue)
        return value



from collections import OrderedDict
from copy import copy
from typing import Literal, cast

from .base import ManagerMethodGenerator


class GeneralMethodGenerator(ManagerMethodGenerator):
    method_name: str = "general"

    def kwargs(
        self, location: Literal["method", "operation"] = "method"
    ) -> OrderedDict[str, str]:
        """
        Just in case this ends up being used for a method that can paginate,
        we'll exclude the pagination arguments.
        """
        args = super().kwargs(location=location)
        if self.client.can_paginate(self.boto3_name):
            _args: OrderedDict[str, str] = OrderedDict()
            for _arg, arg_type in super().kwargs(location=location).items():
                if _arg not in self.PAGINATOR_ARGS:
                    _args[_arg] = arg_type
            return _args
        return args

    @property
    def return_type(self) -> str:
        """
        For generic methods:

        * If :py:attr:`ManagerMethodDefinition.return_type` is set, use that.
        * If :py:attr:`ManagerMethodDefinition.response_attr` is set, infer the
          return type from the response attribute.
        * Otherwise, return the response class name

        Returns:
            The name of the return type class.

        """
        response_class_name = self.response_class
        if self.method_def.return_type:
            return self.method_def.return_type
        if self.response_attr is None:
            return f'"{response_class_name}"'
        if self.output_shape is not None:
            response_attr_shape = self.output_shape.members[
                cast(str, self.response_attr)
            ]
        return_type = self.shape_converter.convert(response_attr_shape, quote=True)
        return return_type  # noqa: RET504

    @property
    def body(self) -> str:
        """
        Generate the method body for a general method.

        .. note::

            This is a complicated one because we have to deal with:

                1. Methods that can paginate.
                2. Methods that return a single response.
                3. Methods that can't paginate but return a list of responses.
                4. Methods that return ``None``

        Returns:
            The method body.

        """

        def generate_paginator_code() -> str:
            code: str = f"""
        paginator = self.client.get_paginator('{self.boto3_name}')
        {self.operation_args}
        response_iterator = paginator.paginate(**{{k: v for k, v in args.items() if v is not None}})
"""  # noqa: E501
            return_type = copy(self.return_type)
            if not return_type.startswith("Optional["):
                return_type = f"Optional[{return_type}]"
            code += f"""
        results: {self.return_type} = []
"""
            if self.response_attr is not None:
                code += f"""
        for _response in response_iterator:
            response = {self.response_class}(**_response)
            if response.{self.response_attr} is not None:
                results.extend(response.{self.response_attr})
            else:
                break
"""
            else:
                code += f"""
        for _response in response_iterator:
            response = {self.response_class}(**_response)
            results.append(response)
"""
            return code

        def handle_response() -> str:
            code = ""
            if self.return_type in ("None", '"None"'):
                return ""
            return_type = copy(self.return_type)
            if not return_type.startswith("Optional["):
                return_type = f"Optional[{return_type}]"
            code += f"""
        results: {self.return_type} = None
        if response is not None:"""
            if self.response_attr is not None:
                code += f"""
            results = response.{self.response_attr}
"""
            else:
                code += """
            results = response
"""
            return code

        if self.client.can_paginate(self.boto3_name):
            code = generate_paginator_code()
        else:
            code = f"""
        {self.operation_args}
        {self.operation_call}
        {handle_response()}
"""
        code += f"""
        self.sessionize(results)
        return cast({self.return_type}, results)
"""
        return code
